#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import, unicode_literals


if __name__ == '__main__':
    import argparse, textwrap
    import script_utils # Append mripy to Python path
    from mripy import dicom_report, utils

    parser = argparse.ArgumentParser(description='',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=textwrap.dedent('''\
            examples:
                1) report_parameters.py ../raw_fmri/func01
                2) report_parameters.py -p mp2rage ../raw_fmri/'T1*'
            '''))
    parser.add_argument('dicom_folder', help='a folder that contains all dicom files for a single sequence')
    parser.add_argument('-p', '--preset', help=f"preset for what parameters to report: {', '.join(dicom_report.preset4sequence.values())}")
    args = parser.parse_args()

    print('>> Reading dicom headers...')
    report, preset, info = dicom_report.report_parameters(args.dicom_folder, preset=args.preset, return_preset=True, return_info=True)
    print(f'>> Using "{preset}" preset')
    print(report)
    print(f">> Coil: {info['TransmittingCoil']},   ReferenceAmplitude: {info['ReferenceAmplitude']} V,   TotalScanTime: {utils.format_duration(info['TotalScanTime'])}")


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import, unicode_literals
from collections import OrderedDict
import numpy as np
from numpy.polynomial import polynomial
from scipy import stats
import pandas as pd
from sklearn import linear_model
from . import six, utils


def nearest(x, parity='odd', round=np.round):
    if parity == 'even':
        return np.int_(round(x/2.0)) * 2
    elif parity == 'odd':
        return np.int_(round((x+1)/2.0)) * 2 - 1


def invert_affine(mat):
    M, v = mat[:,:3], mat[:,3]
    inv = np.linalg.inv(M)
    return np.c_[inv, -inv@v]


def concat_affine(mat2, mat1):
    '''
    mat @ v = mat2 @ mat1 @ v

    For afni's "fetching" convention, the first transform goes first.
    mat1 = io.read_affine('SurfVoltoT1.aff12.1D')   # Fetching SurfVol from T1 grid
    mat2 = io.read_affine('T1toEPI.aff12.1D')       # Fetching T1 from EPI grid
    mat = math.concat_affine(mat1, mat2) # Note the order! The combined transform needs to fetch SurfVol from EPI
    io.write_affine('SurfVoltoEPI.aff12.1D', mat)   # Fetching SurfVol from EPI
    '''
    return np.c_[mat2[:,:3]@mat1[:,:3], mat2[:,:3]@mat1[:,3]+mat2[:,3]]


def apply_affine(mat, xyz):
    '''
    xyz : 3xN array
    '''
    return mat[:,:3] @ xyz + mat[:,3:4]


def LPI2RAI_affine(mat):
    LPI2RAI = np.c_[np.diag([-1,-1,1]), np.zeros(3)] # Interestingly, the inverse of LPI2RAI is itself (i.e., LPI2RAI==RAI2LPI)
    return concat_affine(LPI2RAI, concat_affine(mat, LPI2RAI))

RAI2LPI_affine = LPI2RAI_affine
# fsl2afni_affine = LPI2RAI_affine
# afni2fsl_affine = RAI2LPI_affine


def polyfit3d(x, y, z, f, deg, method=None):
    # https://stackoverflow.com/questions/7997152/python-3d-polynomial-surface-fit-order-dependent
    # https://docs.scipy.org/doc/numpy-1.13.0/reference/routines.polynomials.polynomial.html
    if method is None:
        method = 'ols'
    deg = np.array(deg) if utils.iterable(deg) else np.repeat(deg, 3)
    vander = polynomial.polyvander3d(x, y, z, deg)
    vander = vander.reshape(-1,vander.shape[-1])
    if method == 'ols':
        c = np.linalg.lstsq(vander, f.ravel(), rcond=None)[0]
    elif method == 'ridge':
        model = linear_model.Ridge(fit_intercept=False)
        model.fit(vander, f.ravel())
        c = model.coef_
    elif method == 'lasso':
        model = linear_model.Lasso(fit_intercept=False)
        model.fit(vander, f.ravel())
        c = model.coef_
    c = c.reshape(deg+1)
    return c 


class DomainMapper(object):
    def __init__(self, domain=None):
        self.domain = [0, 2*np.pi] if domain is None else domain

    def to2pi(self, x):
        # Mapping domain to [0, 2*pi]
        y = (x - self.domain[0]) / (self.domain[-1] - self.domain[0]) * 2*np.pi
        return y

    def from2pi(self, y):
        # Mapping domain back to original
        x = y / (2*np.pi) * (self.domain[-1] - self.domain[0]) + self.domain[0]
        return x


def circular_mean(x, domain=None, weight=None, axis=None):
    '''
    Circular mean for values from arbitary circular domain (not necessarily angles).
    '''
    if weight is None:
        weight = np.ones(x.shape)
    # Mapping domain into [0, 2*pi]
    mapper = DomainMapper(domain)
    y = mapper.to2pi(x)
    # Circular mean
    mean_y = np.sum(np.exp(1j*y) * weight, axis=axis) / np.sum(weight, axis=axis)
    mean_y = np.mod(np.angle(mean_y), 2*np.pi)
    # Mapping domain back
    mean_x = mapper.from2pi(mean_y)
    return mean_x


def circular_std(x, domain=None, weight=None, axis=None):
    '''
    Following scipy.stats.circstd()'s definition of circular standard deviation 
    that in the limit of small angles returns the 'linear' standard deviation.
    scipy.stats.circstd() doesn't support weight.
    pycircstat.std() doesn't support domain.
    astropy.stats.circvar() follows another definition.
    '''
    if weight is None:
        weight = np.ones(x.shape)
    # Mapping domain into [0, 2*pi]
    mapper = DomainMapper(domain)
    y = mapper.to2pi(x)
    # Circular std
    mean_y = np.sum(np.exp(1j*y) * weight, axis=axis) / np.sum(weight, axis=axis)
    std_y = np.sqrt(-2*np.log(np.abs(mean_y)))
    # Mapping domain back
    std_x = mapper.from2pi(std_y)
    return std_x


def corrcoef_along_axis(x, y, axis=None):
    x = x - np.mean(x, axis=axis, keepdims=True)
    y = y - np.mean(y, axis=axis, keepdims=True)
    r = np.sum(x * y.conjugate(), axis=axis) / \
        np.sqrt(np.sum(x*x.conjugate(), axis=axis) * np.sum(y*y.conjugate(), axis=axis))
    return r


def circular_corrcoef(x1, x2, domain=None, n_perm=1000, ci=0.95, n_boot=None):
    '''
    The complex corrcoef method used here is fundamentally different from 
    the (Fisher & Lee, 1983) method which is implemented by pycircstat.corrcc(). 
    For more details, read my note
    https://docs.google.com/document/d/1sl39YH3g3TFQu1zX-Ax477NULEyJE--MHMXg2gg0cyk/edit
    '''
    # Mapping domain into [0, 2*pi]
    mapper = DomainMapper(domain)
    y1 = mapper.to2pi(x1)
    y2 = mapper.to2pi(x2)
    # Compute circular corrcoef by treating angles as the phase of complex numbers
    f = lambda a1, a2, axis=None: corrcoef_along_axis(np.exp(1j*a1), np.exp(1j*a2), axis=axis)
    r = f(y1, y2)
    rM, rP = np.abs(r), np.angle(r)
    # Randomization test for significance
    if n_perm:
        n = y1.shape[0]
        Y1 = y1[np.random.choice(np.arange(n), size=[n_perm,n], replace=True)]
        Y2 = y2[np.random.choice(np.arange(n), size=[n_perm,n], replace=True)]
        R = f(Y1, Y2, axis=1)
        p = 1 - stats.percentileofscore(np.abs(R), rM)/100
    else:
        p = None
    # Bootstrap for confidence interval
    if ci is not None:
        if n_boot is None:
            n_boot = n_perm if n_perm else 1000
        idx = np.random.choice(np.arange(n), size=[n_boot,n], replace=True)
        Y1 = y1[idx]
        Y2 = y2[idx]
        R = f(Y1, Y2, axis=1)
        lb = (1-ci)/2
        ub = 1 - lb
        CI = np.percentile(np.abs(R), np.r_[lb, ub]*100)
    else:
        CI = None
    return rM, rP, p, CI


def normalize_logP(logP, axis=None):
    # https://stats.stackexchange.com/questions/66616/converting-normalizing-very-small-likelihood-values-to-probability
    max_logP = np.max(logP, axis=axis, keepdims=True)
    logP = logP - max_logP # Make sure the larger elements (at least the largest one) no longer underflow
    P = np.exp(logP) # It is OK that the smaller elements still underflow
    return P / np.sum(P, axis=axis, keepdims=True)


def gaussian_logpdf(x, mean, cov, cov_inv=None, axis=-1):
    '''
    More efficient multivariate normal distribution log pdf than stats.multivariate_normal.logpdf()
    '''
    z = (x.swapaxes(axis,-1) - mean)[...,np.newaxis] # shape=(...,n,1)
    M = np.linalg.pinv(cov) if cov_inv is None else cov_inv # shape=(n,n). This can be slow, precompute if possible.
    # maxmul() used here is much more efficient (10x) for large matrices than dot() used in stats.multivariate_normal
    maha = (z.swapaxes(-1,-2) @ M @ z)[...,0,0] # (...,1,n) @ (n,n) @ (...,n,1) = (...,1,1)
    return -0.5 * (len(mean)*np.log(2*np.pi) + np.prod(np.linalg.slogdet(cov)) + maha)


def pinv(x):
    # This could be a numpy issue, since the same matrix works fine in Matlab.
    # https://github.com/numpy/numpy/issues/1588
    try:
        res = np.linalg.pinv(x)
    except np.linalg.LinAlgError as err:
        n_retry = 0
        while True:
            try:
                res = np.linalg.pinv(x + np.random.randn(*x.shape)*1e-15)
                break
            except np.linalg.LinAlgError:
                n_retry += 1
                if n_retry == 3:
                    from deepdish import io as dio
                    dio.save('debug.hdf', dict(x=x)) 
                    raise err
    return res


def median_argmax(x, axis=-1):
    y = x.swapaxes(axis, -1)
    res = [np.median(np.nonzero(yy == np.max(yy))[0]) for yy in y.reshape(-1, x.shape[axis])]
    return np.reshape(res, y.shape[:-1]).astype(int)


def tsarray2df(tsarray, t=None, ts_name='value', t_name='time', trial_name='trial', trial_df=None):
    '''
    Parameters
    ----------
    tsarray : ndarray, n_trials * n_times

    Returns
    -------
    df : DataFrame
    '''
    n_trials, n_times = tsarray.shape
    if t is None:
        t = np.arange(n_times)
    tsarray = tsarray.ravel()
    t = np.tile(t, [n_trials, 1]).ravel()
    trials = np.repeat(np.arange(n_trials), n_times)
    df = pd.DataFrame(OrderedDict([(trial_name, trials), (t_name, t), (ts_name, tsarray)]))
    if trial_df is not None:
        df = pd.concat([df, trial_df.iloc[trials].reset_index(drop=True)], axis=1)
    return df


def argsort_rows(rows, cols=None):
    if cols is None:
        cols = np.arange(rows.shape[1])[::-1]
    idx = np.arange(rows.shape[0])
    for col in cols:
        idx = idx[rows[idx,col].argsort(kind='stable')]
    fwd, rev = idx, np.zeros_like(idx)
    rev[fwd] = np.arange(rows.shape[0])
    return fwd, rev


if __name__ == '__main__':
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import, unicode_literals
from pylab import * 


def pylab_eval(expr, **variables):
    return eval(expr, None, variables)


if __name__ == '__main__':
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os.path as path

__author__ = 'herrlich10 <herrlich10@gmail.com>'
with open(path.join(path.dirname(path.realpath(__file__)), '__version__')) as f:
    __version__ = f.readline().strip()


if __name__ == '__main__':
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import inspect, re, glob
from os import path
from collections import OrderedDict
import pandas as pd
from . import dicom


preset4sequence = {
    'epfid2d1': 'ep2d',
    'ep3Dfid3d1': 'ep3d',
    'tfl3d1rr_ns': 'mprage',
    'tfl3d1_ns': 'mp2rage',
    'fl2d1': 'gre',
    'tfi2d1': 'ssfp2d',
}


def report_parameters(dicom_folder, preset=None, return_preset=False, return_info=False):
    # For multi-params sequence (e.g., MP2RAGE)
    if any([w in dicom_folder for w in ['*', '?']]):
        infos = [dicom.parse_series_info(folder) for folder in sorted(glob.glob(dicom_folder))]
    elif all([path.isdir(f) for f in sorted(glob.glob(path.join(dicom_folder, '*')))]):
        infos = [dicom.parse_series_info(folder) for folder in sorted(glob.glob(path.join(dicom_folder, '*')))]
    else: # For ordinary sequence
        infos = [dicom.parse_series_info(dicom_folder)]
    info = infos[0]
    oneline = lambda s: ''.join(inspect.cleandoc(s).split('\n'))
    # Try to automatically determine preset based on info['SequenceName']
    if preset is None:
        for pattern in preset4sequence:
            if re.search(pattern, info['SequenceName']):
                preset = preset4sequence[pattern]
                break
        else:
            raise ValueError(f"Please specify a preset for parameters reporting: {', '.join(preset4sequence.values())}")
    # Phrase
    ND = f"{info['MRAcquisitionType']}"
    TR = f"TR = {info['RepetitionTime']:.0f} ms"
    TE = f"TE = {info['EchoTime']:g} ms"
    FA = f"{info['FlipAngle']:.0f}° flip angle"
    voxel_size = f"voxel size {info['resolution'][0]:.3g} × {info['resolution'][1]:.3g} × {info['resolution'][2]:.3g} mm"
    FOV = f"field of view {info['FOV'][0]:g} × {info['FOV'][1]:g} mm"
    n_slices = f"{info['n_slices']} {info['orientation']} slices"
    bandwidth = f"receiver bandwidth {info['PixelBandwidth']:.0f} Hz/pix"
    PF = (f", {info['PhasePartialFourier']} phase partial Fourier" if info['PhasePartialFourier'] is not None else '') + \
         (f", {info['SlicePartialFourier']} slice partial Fourier" if info['SlicePartialFourier'] is not None else '')
    GRAPPA = (f", GRAPPA acceleration factor {info['GRAPPA']}" if info['GRAPPA'] else '')
    dist_corr = f"{dict(ND='No', DIS2D='2D', DIS3D='3D')[info['distortion_correction']]} distortion correction applied."
    # Sentence
    if preset.lower() in ['ep2d', 'ep3d']:
        effective_TR  = f"TR = {info['TR']*1000/1.004:.0f} ms" # 1.004 is corrected for the fast clock of IBP Siemens 7T
        MB = (f", MultiBand acceleration factor {info['MultiBand']}" if info['MultiBand'] else '')
        report = oneline(f'''T2*-weighted {ND} gradient-echo EPI sequence ({effective_TR}, {TE}, {FA}, {voxel_size}, {FOV}, {n_slices}, {bandwidth}{PF}{GRAPPA}{MB}).''')
    elif preset.lower() in ['mprage']:
        if 'InversionTime' in info:
            TI = f"TI = {info['InversionTime']:.0f} ms"
            report = oneline(f'''T1-weighted {preset.upper()} sequence ({TR}, {TE}, {TI}, {FA}, {voxel_size}, {FOV}, {n_slices}, {bandwidth}{PF}{GRAPPA}).''')
        else:
            report = oneline(f'''PD-weighted {preset.upper()} sequence ({TR}, {TE}, {FA}, {voxel_size}, {FOV}, {n_slices}, {bandwidth}{PF}{GRAPPA}).''')
        report += '\n' + dist_corr
    elif preset.lower() in ['mp2rage']:
        labels = {info_['SeriesDescription'][len(info_['ProtocolName'])+1:]: k for k, info_ in enumerate(infos)}
        TI1 = f"TI1 = {infos[labels['INV1_ND']]['InversionTime']:.0f} ms"
        FA1 = f"{infos[labels['INV1_ND']]['FlipAngle']:.0f}° flip angle"
        TI2 = f"TI2 = {infos[labels['INV2_ND']]['InversionTime']:.0f} ms"
        FA2 = f"{infos[labels['INV2_ND']]['FlipAngle']:.0f}° flip angle"
        report = oneline(f'''T1-weighted {preset.upper()} sequence ({TR}, {TE}, {voxel_size}, {FOV}, {n_slices}, {bandwidth}{PF}{GRAPPA}), yielding two inversion contrasts ({TI1}, {FA1}, {TI2}, {FA2}) which were combined into a single T1-weighted image.''')
    elif preset.lower() in ['gre']:
        report = oneline(f'''T2*-weighted {ND} gradient-echo sequence ({TR}, {TE}, {FA}, {voxel_size}, {FOV}, {n_slices}, {bandwidth}{PF}{GRAPPA}).''')
    elif preset.lower() in ['ssfp2d']:
        report = oneline(f'''{ND} SSFP sequence ({TR}, {TE}, {FA}, {voxel_size}, {FOV}, {n_slices}, {bandwidth}{PF}{GRAPPA}).''')
    res = (report,) + ((preset,) if return_preset else ()) + ((info,) if return_info else ())
    return res if len(res) > 1 else res[0]


def inspect_mp2rage(data_dir, subdir_pattern='T1??'):
    sess_dirs = sorted([f for f in glob.glob(f"{data_dir}/*") if path.isdir(f)])
    df = []
    for sess_dir in sess_dirs:
        T1_folders = glob.glob(f"{sess_dir}/{subdir_pattern}")
        info = dicom.parse_series_info(T1_folders[0])
        df.append(OrderedDict(session=path.basename(sess_dir), resolution='x'.join(f'{d:g}' for d in info['resolution']), 
            ref_amp=info['ReferenceAmplitude'], coil=info['TransmittingCoil'], n_images=len(T1_folders)))
    return pd.DataFrame(df)


if __name__ == '__main__':
    pass


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import, unicode_literals
from os import sys, path

# Make sure the package "mripy" is importable during test time
package_dir = path.abspath(path.join(path.dirname(__file__), '..'))
sys.path.insert(0, path.dirname(package_dir))

# Make sure the test data (which is too big to fit in the package) is available
data_dir = path.abspath(path.join(package_dir, '../mripy_data'))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import, unicode_literals
import sys, argparse, textwrap, shlex, os, shutil, collections
from os import path


def parse_copy(args_copy):
    def is_folder(x):
        return not x[0].isdigit()
    to_copy = collections.OrderedDict()
    if len(args_copy) > 0:
        if not is_folder(args_copy[0]):
            raise ArgumentError("The first parameter following -c is expected to be a folder name, and mustn't start with 0-9.")
        for x in args.copy:
            if is_folder(x):
                folder = x
                to_copy[folder] = []
            else:
                to_copy[folder].append(x)
        for folder, index_list in to_copy.items():
            to_copy[folder] = utils.expand_index_list(index_list, '%04d')
    return to_copy


if __name__ == '__main__':
    import script_utils # Append mripy to Python path
    from mripy import io, utils, afni, dicom

    parser = argparse.ArgumentParser(description='Sort raw dicom files and copy them into separate folders.',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=textwrap.dedent('''\
            examples:
              1) List all datasets in ~/raw_data
                $ sort_dicom.py -l -i ~/raw_data
              2) Copy 0002 to ~/sorted_data/anat, 0003-0006 to ~/sorted_data/func01-04,
                 0008 as func05, 0010 as func06, 0012 as func07
                $ extract_phys.py -i ~/raw_data -o ~/sorted_data -a 2 -f 3-6 8..12(2)\
            '''))
    parser.add_argument('-i', '--input', dest='input_dir', default='.', metavar='path/to/input/dir', help='path/to/input/dir containing raw dicom files, default: pwd')
    parser.add_argument('-o', '--output', dest='output_dir', default=None, help='path/to/output/dir, mkdir if not exists, default: same as input')
    parser.add_argument('-l', '--list', action='store_true', help='list dataset info')
    parser.add_argument('-a', '--anat', nargs='+', default=[], help='anat datasets to copy (e.g., -a 2)')
    parser.add_argument('-f', '--func', nargs='+', default=[], help='func datasets to copy (e.g., -f 3-6 9..15(2))')
    parser.add_argument('-c', '--copy', nargs='+', default=[], help='other datasets to copy (e.g., -c reverse 3 13 forward 4 14)')
    parser.add_argument('-g', '--log', nargs='?', help='write log to (specified) file') # Will be None whether you set -g, unless you specify -g a/new/path
    parser.add_argument('-s', '--select', default=-1, help='select which study to copy, default: last study (-1)')
    parser.add_argument('--pattern', default=io.SERIES_PATTERN, help='regular expression pattern capturing dataset index')
    parser.add_argument('-m', '--method', default='filename', help='method used to sort dicom files: [filename]|header')
    args = parser.parse_args()
    # args, unknowns = parser.parse_known_args()
    if args.output_dir is None:
        args.output_dir = args.input_dir
    if not path.exists(args.output_dir):
        os.makedirs(args.output_dir)
    to_copy = collections.OrderedDict()
    to_copy['anat'] = utils.expand_index_list(args.anat, '%04d')
    to_copy['func'] = utils.expand_index_list(args.func, '%04d')
    to_copy.update(parse_copy(args.copy))
    n_to_copy = sum([len(v) for v in to_copy.values()])
    if args.log is None and n_to_copy > 0:
        args.log = path.join(args.output_dir, 'sort_dicom.log')
    if args.select > 0:
        args.select -= 1 # From one-based index to zero-based index
    # Sort files into studies of series
    if args.method == 'filename':
        studies = io.sort_dicom_series(args.input_dir)
        dicom_parser = io.parse_dicom_header
    elif args.method == 'header':
        print('>> Please wait while extracting headers...')
        studies = dicom.sort_dicom_series(args.input_dir)
        dicom_parser = dicom.parse_dicom_header
    if studies is None:
        print('>> No dicom file found in "{0}". Use -i to specify input dir.'.format(path.realpath(args.input_dir)))
        exit()
    # List all series in each study
    descriptions = []
    for k, study in enumerate(studies):
        descriptions.append(collections.OrderedDict())
        if args.list and len(studies) > 1:
            print('===== study #{0} ====='.format(k+1))
        for sn, files in study.items():
            if afni.has_afni:
                info = io.parse_series_info(files, parser=dicom_parser)
                desc = '{0} ({1}): {2}, {3}{4}'.format(sn,
                    info['n_volumes'] if info['n_volumes']>1 else len(files),
                    info['ProtocolName'],
                    'x'.join(['%.2g' % x for x in info['resolution']]),
                    (', TR=%.2f' % info['TR']) if info['TR'] is not None else '')
            else:
                desc = '{0} ({1})'.format(sn, len(files))
            descriptions[-1][sn] = desc
            if args.list:
                print(desc)
    # Overload print
    if args.log is not None:
        print = script_utils.get_log_printer(args.log)[0]
        print('# {0}'.format(os.getcwd()))
        print('# {0}'.format(' '.join([path.basename(sys.argv[0])] + sys.argv[1:])))
    # Copy files
    for kind, seqs in to_copy.items():
        if len(seqs) > 0:
            print('# {0}'.format(kind))
        for k, sn in enumerate(seqs):
            d = path.join(args.output_dir, '{0}{1:02d}'.format(kind, k+1) if len(seqs) > 1 else kind)
            if not path.exists(d):
                os.makedirs(d)
            print('- {0}'.format(descriptions[args.select][sn]))
            for f in studies[args.select][sn]:
                shutil.copy(f, d)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import, unicode_literals
import subprocess, os
import numpy as np
from collections import OrderedDict
from . import six, afni, io, utils


def afni_costs(base_file, in_file, mask=None):
    if mask is not None:
        emask = utils.temp_prefix(suffix='.nii')
        utils.run(f"3dcalc -a {mask} -expr '1-step(a)' -prefix {emask} -overwrite")
    emask_cmd = f'-emask {emask}' if mask is not None else ''
    cmd = f"3dAllineate -base {base_file} -input {in_file} -allcostX {emask_cmd}"
    lines = afni.check_output(cmd, pattern=r'^\s+\S+\s+=\s+\S+')
    costs = OrderedDict()
    for line in lines:
        name, value = line.split('=')
        name = name.strip()
        if name not in costs:
            costs[name] = []
        costs[name].append(float(value))
    if np.all(np.array([len(v) for k, v in costs.items()]) == 1):
        costs = OrderedDict((k, v[0]) for k, v in costs.items())
    if mask is not None:
        os.remove(emask)
    return costs


def residual_motion(files, print_table=False):
    dsets = [io.read_afni(f) for f in files]
    scores = {'filename': [afni.get_prefix(f) for f in files]}

    # ===== Temporal variation =====
    # Caveates:
    # 1. Across dsets, those of different noise level have different inherent
    #    variability, and thus cannot be compared directly.
    # 2. Within dsets (and across different methods), different levels of
    #    spatial smoothing (caused by resampling) also affect this measure.
    scores['tSTD'] = [np.mean(np.std(dset, axis=-1)) for dset in dsets]

    # ===== Motion as simultaneous intensity change across many voxels =====
    # x = np.zeros((len(dsets), dsets[0].shape[-1]-1))
    # for k, dset in enumerate(dsets):
    #     d = np.diff(dset, axis=-1)
    #     np.abs()
    #     x[k,:] = np.mean(, axis=(0,1,2))


    # Print tabular results using pandas
    if print_table:
        import pandas as pd
        df = pd.DataFrame(scores)
        print(df)


if __name__ == '__main__':
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2018 herrlich10@gmail.com
#
# Permission is hereby granted, free of charge, to any person obtaining a copy 
# of this software and associated documentation files (the "Software"), to deal 
# in the Software without restriction, including without limitation the rights 
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
# copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included 
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL 
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
# SOFTWARE.

from __future__ import print_function, division, absolute_import, unicode_literals
import sys, os, shlex, time, textwrap, re
import subprocess, multiprocessing, queue, threading, ctypes, uuid
import numpy as np

__author__ = 'herrlich10 <herrlich10@gmail.com>'
__version__ = '0.1.7'

# The following are copied from six
# =================================
if sys.version_info[0] == 3:
    string_types = (str,)
    from io import StringIO
else:
    string_types = (basestring,)
    from StringIO import StringIO

def add_metaclass(metaclass):
    """Class decorator for creating a class with a metaclass."""
    def wrapper(cls):
        orig_vars = cls.__dict__.copy()
        slots = orig_vars.get('__slots__')
        if slots is not None:
            if isinstance(slots, str):
                slots = [slots]
            for slots_var in slots:
                orig_vars.pop(slots_var)
        orig_vars.pop('__dict__', None)
        orig_vars.pop('__weakref__', None)
        return metaclass(cls.__name__, cls.__bases__, orig_vars)
    return wrapper
# =================================


def format_duration(duration, format='standard'):
    '''Format duration (in seconds) in a more human friendly way.
    '''
    if format == 'short':
        units = ['d', 'h', 'm', 's']
    elif format == 'long':
        units = [' days', ' hours', ' minutes', ' seconds']
    else: # Assume 'standard'
        units = [' day', ' hr', ' min', ' sec']
    values = [int(duration//86400), int(duration%86400//3600), int(duration%3600//60), duration%60]
    for K in range(len(values)): # values[K] would be the first non-zero value
        if values[K] > 0:
            break
    formatted = ((('%d' if k<len(values)-1 else '%.3f') % values[k]) + units[k] for k in range(len(values)) if k >= K)
    return ' '.join(formatted)


def cmd_for_exec(cmd, shell=False):
    ''' Format cmd appropriately for execution according to whether shell=True.

    Split a cmd string into a list, if shell=False.
    Join a cmd list into a string, if shell=True.
    Do nothing to callable.

    Parameters
    ----------
    cmd : str, list, or callable
    shell : bool
    '''
    # If shell=kwargs, its true value is inferred.
    if isinstance(shell, dict):
        shell = ('shell' in shell and shell['shell'])
    if not callable(cmd):
        if shell: # cmd string is required
            if not isinstance(cmd, string_types):
                cmd = ' '.join(cmd)
        else: # cmd list is required
            if isinstance(cmd, string_types):
                cmd = shlex.split(cmd) # Split by space, preserving quoted substrings
    return cmd


def cmd_for_disp(cmd):
    '''Format cmd for printing.
    
    Parameters
    ----------
    cmd : str, list, or callable
    '''
    if not callable(cmd):
        if isinstance(cmd, string_types):
            cmd = shlex.split(cmd) # Remove insignificant whitespaces
        cmd = ' '.join(shlex.quote(s) for s in cmd)
    return cmd


ERROR_PATTERN = r'error|^\*{2}\s'


def check_output_for_errors(output, error_pattern=None, verbose=1, label=''):
    '''
    User can skip error checking by setting error_pattern=''
    '''
    if error_pattern is None:
        error_pattern = ERROR_PATTERN
    n_errors = 0
    if error_pattern != '': # User can skip error checking by setting error_pattern=''
        if isinstance(error_pattern, string_types): # User can provide compiled regex if case sensitivity is desired
            error_pattern = re.compile(error_pattern, re.IGNORECASE)
        for line in output:
            if error_pattern.search(line):
                if verbose > 0:
                    print(label, line, end='')
                n_errors += 1
    return n_errors


def check_output_for_goal(output, goal_pattern=None):
    if goal_pattern is None:
        return True
    if isinstance(goal_pattern, string_types): # User can provide compiled regex if case sensitivity is desired
        goal_pattern = re.compile(goal_pattern, re.IGNORECASE)
    for line in output:
        if goal_pattern.search(line):
            return True
    return False


def run(cmd, check=True, error_pattern=None, goal_pattern=None, shell=False, verbose=2):
    '''Run an external command line.
    
    This function is similar to subprocess.run introduced in Python 3.5, but
    provides a slightly simpler and perhaps more convenient API.

    Parameters
    ----------
    cmd : str or list
    '''
    cmd = cmd_for_exec(cmd, shell=shell)
    cmd_str = cmd_for_disp(cmd)
    if verbose > 0:
        print('>>', cmd_str)
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=shell)
    res = {'cmd': cmd_str, 'pid': p.pid, 'output': [], 'start_time': time.time()}
    for line in iter(p.stdout.readline, b''): # The 2nd argument is sentinel character (there will be no ending empty line)
        res['output'].append(line.decode('utf-8'))
        if verbose > 1:
            print(res['output'][-1], end='')
    p.stdout.close() # Notify the child process that the PIPE has been broken
    res['returncode'] = p.wait()
    res['stop_time'] = time.time()
    if verbose > 0:
        print('>> Command finished in {0}.'.format(format_duration(res['stop_time'] - res['start_time'])))
    if check and (res['returncode'] or check_output_for_errors(res['output'], error_pattern=error_pattern, verbose=verbose)):
        print('>> Please pay attention to the above errors.')
        raise RuntimeError(f'Error occurs when executing the following command (returncode={p.returncode}):\n{cmd_str}')
    if check and not check_output_for_goal(res['output'], goal_pattern=goal_pattern):
        raise RuntimeError(f'Expected goal pattern "{goal_pattern}" does not found! Something must be wrong!')
    return res


STDOUT = sys.stdout
STDERR = sys.stderr

class TeeOut(StringIO):
    def __init__(self, err=False, tee=True):
        super().__init__()
        self.err = err
        self.tee = tee

    def write(self, s):
        super().write(s)
        if self.err: # Always output error message
            STDERR.write(s)
        elif self.tee:
            STDOUT.write(s)


class PooledCaller(object):
    '''
    Execute multiple command line programs, as well as python callables, 
    asynchronously and parallelly across a pool of processes.
    '''
    def __init__(self, pool_size=None, verbose=1):
        if pool_size is None:
            self.pool_size = multiprocessing.cpu_count() * 3 // 4
        else:
            self.pool_size = pool_size
        self.verbose = verbose
        self.ps = []
        self.cmd_queue = [] # Queue for commands and callables, as well as any additional args
        self._n_cmds = 0 # Auto increased counter for generating cmd idx
        self._idx2pid = {}
        self._pid2job = {} # Hold all jobs for each wait()
        self._log = [] # Hold all jobs across waits (entire execution history for this PooledCaller instance)
        self._fulfilled = {} # Fulfilled dependencies across waits (a faster API compared with self._log)
        self.res_queue = multiprocessing.Queue() # Queue for return values of executed python callables
 
    def run(self, cmd, *args, _depends=None, _retry=None, _dispatch=False, _error_pattern=None, _suppress_warning=False, _block=False, **kwargs):
        '''Asynchronously run command or callable (queued execution, return immediately).
        
        See subprocess.Popen() for more information about the arguments.

        Multiple commands can be separated with ";" and executed sequentially 
        within a single subprocess in linux/mac, only if shell=True.
        
        Python callable can also be executed in parallel via multiprocessing.
        Note that although return values of the callable are retrieved via PIPE,
        sometimes it could be advantageous to directly save the computation 
        results into a shared file (e.g., an HDF5 file), esp. when they're large.
        In the later case, a proper lock mechanism via multiprocessing.Lock() 
        is required.

        Parameters
        ----------
        cmd : list, str, or callable
            Computation in command line programs is handled with subprocess.
            Computation in python callable is handled with multiprocessing.
        shell : bool
            If provided, must be a keyword argument.
            If shell is True, the command will be executed through the shell.
        *args, **kwargs : 
            If cmd is a callable, *args and **kwargs are passed to the callable as its arguments.
            If cmd is a list or str, **kwargs are passed to subprocess.Popen().
        _depends : list
            A list of jobs (identified by their uuid) that have to be done 
            before this job can be scheduled.
        _retry: int
            Number of retry before accepting failure (if detecting non-zero return code).
        _dispatch : bool
            Dispatch the job immediately, which will run in the background without blocking.
        _error_pattern : str
        _suppress_warning : bool
        _block : bool
            if True, call wait() internally and block.

        Returns
        -------
        _uuid : str
            The uuid of current job (which can be used as future jobs' dependency)
        '''
        cmd = cmd_for_exec(cmd, shell=kwargs)
        _uuid = uuid.uuid4().hex[:8]
        if _retry is None:
            _retry = 0
        self.cmd_queue.append((self._n_cmds, cmd, args, kwargs, _uuid, _depends, _retry, _error_pattern, _suppress_warning))
        self._n_cmds += 1 # Accumulate by each call to run(), and reset after wait()
        if _dispatch:
            self.dispatch()
        if _block:
            self.wait()
        return _uuid

    def run1(self, cmd, *args, _error_pattern=None, _suppress_warning=False, **kwargs):
        self.run(cmd, *args, _error_pattern=_error_pattern, _suppress_warning=_suppress_warning, **kwargs)
        return self.wait()

    def _callable_wrapper(self, idx, cmd, *args, **kwargs):
        out = TeeOut(tee=(self.verbose > 1))
        err = TeeOut(err=True)
        sys.stdout = out # This substitution only affect spawned process
        sys.stderr = err
        res = None # Initialized in case of exception
        try:
            res = cmd(*args, **kwargs)
        except Exception as e:
            print('>> Error occurs in job#{0}'.format(idx), file=err)
            print('** ERROR:', e, file=err) # AFNI style error message
            raise e # Re-raise and let parent process to handle it
        finally:
            # Grab all output at the very end of the process (assume that there aren't too much of them)
            # TODO: This could be a potential bug...
            # https://ryanjoneil.github.io/posts/2014-02-14-capturing-stdout-in-a-python-child-process.html
            output = out.getvalue().splitlines(True) + err.getvalue().splitlines(True)
            self.res_queue.put([idx, res, output]) # Communicate return value and output (Caution: The underlying pipe has limited size. Have to get() soon in wait().)

    def _async_reader(self, idx, f, output_list, speed_up, suppress_warning=False):
        while True: # We can use event to tell the thread to stop prematurely, as demonstrated in https://stackoverflow.com/questions/323972/is-there-any-way-to-kill-a-thread
            line = f.readline()
            line = line.decode('utf-8')
            if line: # This is not lock protected, because only one thread (i.e., this thread) is going to write
                output_list.append(line)
                if (line.startswith('*') or line.startswith('\x1b[7m')) and not suppress_warning: # Always print AFNI style WARNING and ERROR through stderr unless explicitly suppressed
                    # '\x1b[7m' and '\x1b[0m' are 'reverse' and 'reset' respectively (https://gist.github.com/abritinthebay/d80eb99b2726c83feb0d97eab95206c4)
                    print('>> Something happens in job#{0}'.format(idx), file=sys.stderr)
                    print(line, end='', file=sys.stderr)
                elif self.verbose > 1:
                    print(line, end='')
            else: # Empty line signifies the end of the spawned process
                break
            if not speed_up.is_set():
                time.sleep(0.1) # Don't need to poll for output too aggressively during run time

    def dispatch(self):
        # If there are free slot and more jobs
        # while len(self.ps) < self.pool_size and len(self.cmd_queue) > 0:
        if len(self.ps) < self.pool_size and len(self.cmd_queue) > 0:
            idx, cmd, args, kwargs, _uuid, _depends, _retry, _error_pattern, _suppress_warning = self.cmd_queue.pop(0)
            if _depends is None or all([dep in self._fulfilled for dep in _depends]): # No dependency or all fulfilled
                # Create a job process only after it is popped from the queue
                job = {'idx': idx, 'cmd': cmd, 'args': args, 'kwargs': kwargs, 'uuid':  _uuid, 
                    'depends': _depends, 'retry': _retry, 'error_pattern': _error_pattern , 
                    'suppress_warning': _suppress_warning, 'output': []} 
                if self.verbose > 0:
                    print('>> job#{0}: {1}'.format(idx, cmd_for_disp(job['cmd'])))
                if callable(cmd):
                    # TODO: Add an if-else branch here if shared memory doesn't work for wrapper 
                    p = multiprocessing.Process(target=self._callable_wrapper, args=(idx, cmd) + args, kwargs=kwargs)
                    p.start()
                else:
                    # Use PIPE to capture output and error message
                    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, **kwargs)
                    # Capture output without blocking (the main thread) by using a separate thread to do the blocking readline()
                    job['speed_up'] = threading.Event()
                    job['watcher'] = threading.Thread(target=self._async_reader, args=(idx, p.stdout, 
                        job['output'], job['speed_up'], job['suppress_warning']), daemon=True)
                    job['watcher'].start()
                self.ps.append(p)
                job['start_time'] = time.time()
                job['pid'] = p.pid
                job['successor'] = None
                job['log_idx'] = len(self._log)
                self._idx2pid[idx] = p.pid
                self._pid2job[p.pid] = job
                self._log.append(job)
            else: # Re-queue the job whose dependencies are not fully fulfilled to the END of the queue
                self.cmd_queue.append((idx, cmd, args, kwargs, _uuid, _depends, _retry, _error_pattern, _suppress_warning))

    def _async_get_res(self, res_list):
        try:
            res = self.res_queue.get(block=False) # idx, return_value, output
        except queue.Empty:
            pass
        else:
            res_list.append(res[:2])
            if len(res) > 2: # For callable only
                job = self._pid2job[self._idx2pid[res[0]]]
                job['output'] = res[2]

    def wait(self, pool_size=None, return_codes=False, return_jobs=False):
        '''
        Wait for all jobs in the queue to finish.
        
        Returns
        -------
        return_values : list
            Return values of executed python callable. Always `None` for command.
        codes : list (only when return_codes=True)
            The return code of the child process for each job.
        jobs : list (only when return_jobs=True)
            Detailed information about each child process, including captured stdout and stderr.
        '''
        if isinstance(pool_size, string_types) and pool_size == 'balanced':
            # Make sure each volley has roughly equal number of jobs
            n = len(self.cmd_queue)
            pool_size = int(np.ceil(n/np.ceil(n/self.pool_size)))
        if pool_size is not None:
            # Allow temporally adjust pool_size for current batch of jobs
            old_size = self.pool_size
            self.pool_size = pool_size
        start_time = time.time()
        ress = []
        while len(self.ps) > 0 or len(self.cmd_queue) > 0:
            # Dispatch jobs if possible
            self.dispatch()
            # Poll workers' state
            for p in self.ps:
                job = self._pid2job[p.pid]
                if isinstance(p, subprocess.Popen):
                    if p.poll() is not None: # If the process is terminated
                        job['stop_time'] = time.time()
                        job['returncode'] = p.returncode
                        job['speed_up'].set()
                        job['watcher'].join() # Retrieve all remaining output before closing PIPE
                        p.stdout.close() # Notify the child process that the PIPE has been broken
                        self.ps.remove(p)
                        if self.verbose > 0:
                            print('>> job#{0} finished (return {1}) in {2}.'.format(job['idx'], job['returncode'], format_duration(job['stop_time']-job['start_time'])))
                        if job['returncode'] != 0: # Failed
                            if job['retry'] > 0: # Need retry
                                # Insert a new cmd (as if we automatically run it again)
                                self.cmd_queue.append((self._n_cmds, job['cmd'], job['args'], job['kwargs'], job['uuid'], 
                                    job['depends'], job['retry']-1, job['error_pattern'], job['suppress_warning']))
                                job['successor'] = self._n_cmds
                                self._n_cmds += 1
                            else: # No more retry, accept failure...
                                raise RuntimeError(f">> job#{job['idx']} failed!\n Full output:\n {''.join(job['output'])}")
                        else: # Successful
                            self.res_queue.put([job['idx'], None]) # Return None to mimic callable behavior
                            self._fulfilled[job['uuid']] = job['log_idx'] # Marked as fulfilled, even with error (TODO: or shall I break all??)
                        # These helper objects may not be useful for the end users
                        for key in ['watcher', 'speed_up', 'args', 'kwargs']:
                            job.pop(key) 
                    else:
                        pass
                elif isinstance(p, multiprocessing.Process):
                    if not p.is_alive(): # If the process is terminated
                        job['stop_time'] = time.time()
                        job['returncode'] = p.exitcode # subprocess.Popen and multiprocessing.Process use different names for this
                        self.ps.remove(p)
                        if self.verbose > 0:
                            print('>> job#{0} finished (return {1}) in {2}.'.format(job['idx'], job['returncode'], format_duration(job['stop_time']-job['start_time'])))
                        # TODO: retry mechanism for callable
                        self._fulfilled[job['uuid']] = job['log_idx'] # Marked as fulfilled
                        # Remove potentially very large data
                        for key in ['args', 'kwargs']:
                            job.pop(key) 
                    else:
                        pass
            time.sleep(0.1)
            # Dequeuing, see https://stackoverflow.com/questions/10028809/maximum-size-for-multiprocessing-queue-item
            self._async_get_res(ress)
        # Handle return values by callable cmd
        while not self.res_queue.empty():
            self._async_get_res(ress)
        ress = [res[1] for res in sorted(ress, key=lambda res: res[0])]
        # Handle return codes by children processes
        jobs = sorted([job for job in self._pid2job.values() if job['successor'] is None], key=lambda job: job['idx'])
        codes = [job['returncode'] for job in jobs]
        if self.verbose > 0:
            duration = time.time() - start_time
            print('>> All {0} jobs done in {1}.'.format(self._n_cmds, format_duration(duration)))
            if np.any(codes):
                print('returncodes: {0}'.format(codes))
                first_error = np.nonzero(codes)[0][0]
                print(f">> Output for job#{first_error} was as follows:\n------------------------------")
                print(jobs[first_error]['output'])
            else:
                print('all returncodes are 0.')
            if self.all_successful(jobs=jobs):
                print('>> All {0} jobs finished successfully.'.format(len(jobs)))
            else:
                print('>> Please pay attention to the above errors.')
        # Reset object states
        self._n_cmds = 0
        self._idx2pid = {}
        self._pid2job = {}
        if pool_size is not None:
            self.pool_size = old_size
        res = (ress,) + ((codes,) if return_codes else ()) + ((jobs,) if return_jobs else ())
        if len(res) == 1:
            return res[0]
        else:
            return res

    def all_successful(self, jobs=None, verbose=None):
        if jobs is None:
            jobs = self._log
        if verbose is None:
            verbose = self.verbose
        # Check return codes
        all_zero = not np.any([job['returncode'] for job in jobs])
        # Check output
        n_errors = sum([check_output_for_errors(job['output'], error_pattern=job['error_pattern'], verbose=verbose, label='[job#{0}]'.format(job['idx'])) for job in jobs])
        return all_zero and n_errors == 0

    def idss(self, total, batch_size=None):
        if batch_size is None:
            batch_size = int(np.ceil(total / self.pool_size / 10))
        return (range(k, min(k+batch_size, total)) for k in range(0, total, batch_size))

    def __call__(self, job_generator, **kwargs):
        # This is similar to the joblib.Parallel signature, which is the only way to
        # pass both args and kwargs for inner execution.
        # >>> pc(pc.run(f"3dvolreg -prefix ... {func}{run}.nii") for run in runs)
        # 
        # It also allows each call to deal with a batch of jobs for better performance, 
        # if the callable is purposely designed to do so, which is especially useful 
        # when there are a huge amount of small jobs.
        # >>> pc(pc.run(compute_depth, ids, *args) for ids in pc.idss(len(depths)))
        n_jobs = 0
        for _ in job_generator: # Queue all jobs from the generator
            n_jobs += 1
        if self.verbose > 0:
            print('>> Start with a total of {0} jobs...'.format(n_jobs))
        return self.wait(**kwargs) # Wait all jobs to finish


class ArrayWrapper(type):
    '''
    This is the metaclass for classes that wrap an np.ndarray and delegate 
    non-reimplemented operators (among other magic functions) to the wrapped array.
    '''
    def __init__(cls, name, bases, dct):
        def make_descriptor(name):
            '''
            Implementation notes
            --------------------
            1. Method (or non-data) descriptors are objects that define __get__() method
               but not __set__() method. Refer to [here](https://docs.python.org/3.6/howto/descriptor.html).
            2. The magic methods of an object (e.g., arr.__add__) are descriptors, not callable.
               So here we must return a property (with getter only), not a lambda.
            3. Strangely, the whole thing must be wrapped in a nested function. See [here](
               https://stackoverflow.com/questions/9057669/how-can-i-intercept-calls-to-pythons-magic-methods-in-new-style-classes).
            4. The wrapped array must be named self.arr
            '''
            return property(lambda self: getattr(self.arr, name))

        type.__init__(cls, name, bases, dct)
        ignore = 'class mro new init setattr getattr getattribute'
        ignore = set('__{0}__'.format(name) for name in ignore.split())
        for name in dir(np.ndarray):
            if name.startswith('__'):
                if name not in ignore and name not in dct:
                    setattr(cls, name, make_descriptor(name))


@add_metaclass(ArrayWrapper) # Compatibility code from six
class SharedMemoryArray(object):
    '''
    This class can be used as a usual np.ndarray, but its data buffer
    is allocated in shared memory (under Cached Files in memory monitor), 
    and can be passed across processes without any data copy/duplication, 
    even when write access happens (which is lock-synchronized).

    The idea is to allocate memory using multiprocessing.Array, and  
    access it from current or another process via a numpy.ndarray view, 
    without actually copying the data.
    So it is both convenient and efficient when used with multiprocessing.

    This implementation also demonstrates the power of composition + metaclass,
    as opposed to the canonical multiple inheritance.
    '''
    def __init__(self, dtype, shape, initializer=None, lock=True):
        self.dtype = np.dtype(dtype)
        self.shape = shape
        if initializer is None:
            # Preallocate memory using multiprocessing is the preferred usage
            self.shared_arr = multiprocessing.Array(self.dtype2ctypes[self.dtype], int(np.prod(self.shape)), lock=lock)
        else:
            self.shared_arr = multiprocessing.Array(self.dtype2ctypes[self.dtype], initializer, lock=lock)
        if not lock:
            self.arr = np.frombuffer(self.shared_arr, dtype=self.dtype).reshape(self.shape)
        else:
            self.arr = np.frombuffer(self.shared_arr.get_obj(), dtype=self.dtype).reshape(self.shape)
 
    @classmethod
    def zeros(cls, shape, dtype=float, lock=True):
        '''
        Return a new array of given shape and dtype, filled with zeros.

        This is the preferred usage, which avoids holding two copies of the
        potentially very large data simultaneously in the memory.
        '''
        return cls(dtype, shape, lock=lock)

    @classmethod
    def from_array(cls, arr, lock=True):
        '''
        Initialize a new shared-memory array with an existing array.
        '''
        # return cls(arr.dtype, arr.shape, arr.ravel(), lock=lock) # Slow and memory inefficient, why?
        a = cls.zeros(arr.shape, dtype=arr.dtype, lock=lock)
        a[:] = arr # This is a more efficient way of initialization
        return a

    def __getattr__(self, attr):
        if attr in self._SHARED_ARR_ATTRIBUTES:
            return getattr(self.shared_arr, attr)
        else:
            return getattr(self.arr, attr)

    def __dir__(self):
        return list(self.__dict__.keys()) + self._SHARED_ARR_ATTRIBUTES + dir(self.arr)

    _SHARED_ARR_ATTRIBUTES = ['acquire', 'release', 'get_lock']

    # At present, only numerical dtypes are supported.
    dtype2ctypes = {
        bool: ctypes.c_bool,
        int: ctypes.c_long,
        float: ctypes.c_double,
        np.dtype('bool'): ctypes.c_bool,
        np.dtype('int64'): ctypes.c_long,
        np.dtype('int32'): ctypes.c_int,
        np.dtype('int16'): ctypes.c_short,
        np.dtype('int8'): ctypes.c_byte,
        np.dtype('uint64'): ctypes.c_ulong,
        np.dtype('uint32'): ctypes.c_uint,
        np.dtype('uint16'): ctypes.c_ushort,
        np.dtype('uint8'): ctypes.c_ubyte,
        np.dtype('float64'): ctypes.c_double,
        np.dtype('float32'): ctypes.c_float,
        }
        



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import, unicode_literals
import unittest
from mripy import utils


class test_utils(unittest.TestCase):
    def test_fname_with_ext(self):
        self.assertEqual(utils.fname_with_ext('prefix+orig', '.HEAD'), 'prefix+orig.HEAD')
        self.assertEqual(utils.fname_with_ext('prefix', '+orig.HEAD'), 'prefix+orig.HEAD')
        self.assertEqual(utils.fname_with_ext('prefix+orig', '+orig.HEAD'), 'prefix+orig.HEAD')
        self.assertEqual(utils.fname_with_ext('prefix+orig.', '+orig.HEAD'), 'prefix+orig.HEAD')
        self.assertEqual(utils.fname_with_ext('prefix+orig.HEAD', '+orig.HEAD'), 'prefix+orig.HEAD')

if __name__ == '__main__':
    unittest.main()



import setuptools
import mripy

with open('README.rst', 'r') as f:
    long_description = f.read()

setuptools.setup(
   name='mripy',
   version=mripy.__version__,
   author='herrlich10',
   author_email='herrlich10@gmail.com',
   description='High resolution fMRI data analysis in Python, based on AFNI, FreeSurfer, ANTs, and many other tools.',
   long_description=long_description,
   long_description_content_type='text/x-rst',
   url='https://github.com/herrlich10/mripy',
   packages=setuptools.find_packages(),
   include_package_data=True,
   install_requires=[
       # 'numpy',
   ],
   classifiers=(
       'Programming Language :: Python :: 3',
       'Operating System :: OS Independent',
       'License :: OSI Approved :: MIT License',
       'Intended Audience :: Developers',
       'Intended Audience :: Science/Research',
       'Topic :: Scientific/Engineering :: Information Analysis',
       'Topic :: Software Development :: Libraries',
       'Topic :: Utilities',
   ),
   python_requires='>=3.6',
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import, unicode_literals
import argparse, textwrap
import numpy as np
from matplotlib import pyplot as plt


if __name__ == '__main__':
    import script_utils # Append mripy to Python path
    from mripy import vis

    parser = argparse.ArgumentParser(description='',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=textwrap.dedent('''\
            examples:
              1) Plot head motion across multiple runs
                $ 1dplot.py -vr dfile.epi*.1D
            '''))
    parser.add_argument('files', nargs='+', help='1D file(s)')
    parser.add_argument('-vr', '--volreg', action='store_true', help='plot head motion parameters')
    parser.add_argument('-y', '--ylim', metavar='L U', nargs='+', help='set ylim')
    parser.add_argument('-s', '--save', metavar='FNAME', help='save figure as *.png')
    parser.add_argument('--silent', action='store_true', help='do not show figure')
    args = parser.parse_args()

    if args.volreg:
        vis.plot_volreg(args.files)
    else:
        x = np.loadtxt(args.files[0])
        plt.plot(x)
    if args.ylim is not None:
        if len(args.ylim) == 1:
            # In case the two parameters are quoted together
            # Yeah, this is a little bit too much...
            plt.ylim(np.float_(args.ylim[0].split()))
        else:
            plt.ylim(np.float_(args.ylim))
    if args.save:
        plt.savefig(args.save)
    if not args.silent:
        plt.show()
        plt.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import, unicode_literals
import unittest
import copy
import numpy as np
from mripy import timecourse


class test_Attributes(unittest.TestCase):
    def setUp(self):
        self.attr = timecourse.Attributes(shape=[6,4])
        self.attr.add('tid', range(6), axis=0)
        self.attr.add('eye', [1, 1, -1, -1], axis=1)
    
    def test_copy(self):
        attr1 = copy.copy(self.attr)
        self.assertNotEqual(id(attr1.attributes), id(self.attr.attributes))
        self.assertNotEqual(id(attr1.attributes['tid']), id(self.attr.attributes['tid']))
        self.assertNotEqual(id(attr1.eye), id(self.attr.eye))
        self.assertTrue(np.all(attr1.tid==self.attr.tid))

    def test_pick(self):
        attr1 = self.attr.pick(slice(0,3), axis=0)
        self.assertNotEqual(id(attr1), id(self.attr))
        self.assertNotEqual(id(attr1.attributes['tid']), id(self.attr.attributes['tid']))
        self.assertNotEqual(id(attr1.eye), id(self.attr.eye))
        self.assertTrue(np.all(attr1.tid==[0, 1, 2]))
        self.assertTrue(np.all(attr1.eye==self.attr.eye))
        attr2 = self.attr.pick([slice(3,None), [True, False, True, False]], axis=[0, 1])
        self.assertTrue(np.all(attr2.tid==[3, 4, 5]))
        self.assertTrue(np.all(attr2.eye==[1, -1]))


class test_Epochs_Attributes(unittest.TestCase):
    def setUp(self):
        self.epochs = timecourse.Epochs.from_array(np.random.rand(6,4,5), TR=2)
        self.epochs.add_event_attr('tid', range(6))
        self.epochs.add_feature_attr('eye', [1, 1, -1, -1])
    
    def test_copy(self):
        epochs1 = self.epochs.copy()
        self.assertEqual(id(epochs1.data), id(self.epochs.data))
        self.assertNotEqual(id(epochs1.attr), id(self.epochs.attr))
        self.assertNotEqual(id(epochs1.attr.eye), id(self.epochs.attr.eye))
    
    def test_pick(self):
        epochs1 = self.epochs[::2]
        self.assertNotEqual(id(epochs1.attr), id(self.epochs.attr))
        self.assertNotEqual(id(epochs1.attr.attributes['eye']), id(self.epochs.attr.attributes['eye']))
        self.assertNotEqual(id(epochs1.attr.tid), id(self.epochs.attr.tid))
        self.assertTrue(np.all(epochs1.attr.tid==[0, 2, 4]))
        self.assertTrue(np.all(epochs1.attr.eye==self.epochs.attr.eye))
        self.assertTrue(np.all(epochs1.data==self.epochs.data[::2]))


# unittest.main(argv=['ignored', '-v'], exit=False) # 'ignored' is required, '-v' is verbose

if __name__ == '__main__':
    unittest.main()
